@@ -12,8 +12,6 @@ Commands:
   write
   record
   create
-  create-dataport
-  create-client
   update
   map
   unmap
@@ -23,17 +21,15 @@ Commands:
   info
   flush
   tree
-  lookup-rid
-  drop-all-children
   script
 
 Options:
-  --host=<host>        OneP URL. Default is $EXO_HOST or m2.exosite.com.
-  --httptimeout=<sec>  HTTP timeout setting.
+  --host=<host>        OneP URL. Default is $EXO_HOST or m2.exosite.com
+  --httptimeout=<sec>  HTTP timeout
   --https              Enable HTTPS
   --debug              Show info like stack traces
-  -h --help            Show this screen.
-  -v --version         Show version.
+  -h --help            Show this screen
+  -v --version         Show version
 
 See 'exo <command> -h' for more information on a specific command.
 """
@@ -49,11 +45,16 @@ from datetime import datetime
 import time
 from pprint import pprint
 from collections import OrderedDict
+import itertools
+import math
 
 from docopt import docopt
 from onepv1lib import onep
 from onepv1lib import onep_exceptions
-from exoline import __version__
+try:
+    from ..exoline import __version__
+except:
+    from exoline import __version__
 
 DEFAULT_HOST='m2.exosite.com'
 cmd_doc = {
@@ -62,10 +63,12 @@ cmd_doc = {
     exo [options] read <cik> [<rid>]
 
 Options:
-    --follow          continue reading
-    --limit=<limit>   limit to [default: 1]
-    --selection=all|autowindow|givenwindow  how to filter results [default: all]
-    --format=raw|csv output format [default: csv]''',
+    --follow                 continue reading
+    --limit=<limit>          number of data points to read [default: 1]
+    --selection=all|autowindow|givenwindow  downsample method [default: all]
+    --format=csv|raw         output format [default: csv]
+    --timeformat=unix|human  unix timestamp or human-readable? [default: human]
+    --intervals              show distribution of intervals between points''',
     'write':
         '''Write data at the current time.\n\nUsage:
     exo [options] write <cik> [<rid>] --value=<value>''',
@@ -74,25 +77,32 @@ Options:
     exo [options] record <cik> [<rid>] ((--value=<timestamp,value> ...) | -)
     exo [options] record <cik> [<rid>] --interval=<seconds> ((--value=<value> ...) | -)
 
-    Pass - to read data from stdin.
-    Pass --interval to generate timestamps at a regular interval from now.
-    If --interval is positive, data will be placed in the future. If it's
-    negative, data will be placed in the past.''',
+    - reads data from stdin.
+    --interval generates timestamps at a regular interval into the past.''',
     'create':
-        '''Create a resource from a json description passed on stdin.\n\nUsage:
-    exo [options] create <cik> (--type=client|clone|dataport|datarule|dispatch) -''',
+        '''Create a resource from a json description passed on stdin, or using
+    defaults.\n\nUsage:
+    exo [options] create <cik> (--type=client|clone|dataport|datarule|dispatch) -
+    exo [options] create <cik> --type=client
+    exo [options] create <cik> --type=dataport (--format=binary|boolean|float|integer|string)
+
+Options:
+    --name=<name     set a resource name (overwriting the one in stdin if present)
+    --alias=<alias>  set an alias
+    --ridonly        just output the RID by itself on a line
+
+Details:
+    Pass - and a json description object on stdin for maximum control.
+    Description is documented here:
+    http://developers.exosite.com/display/OP/Remote+Procedure+Call+API#RemoteProcedureCallAPI-create
+
+    If - is not present, creates a resource with common defaults.''',
     'listing':
         '''List a client's children based on their type.\n\nUsage:
     exo [options] listing <cik> (--type=client|dataport|datarule|dispatch) ... [--plain] [--pretty]''',
     'info':
         '''Get info for a resource in json format.\n\nUsage:
     exo [options] info <cik> [<rid>] [--cikonly] [--pretty]''',
-    'create-dataport':
-        '''Create a dataport.\n\nUsage:
-    exo [options] create-dataport <cik> (--format=binary|boolean|float|integer|string) [--name=<name>]''',
-    'create-client':
-        '''Create a client.\n\nUsage:
-    exo [options] create-client <cik> [--name=<name>]''',
     'update':
         '''Update a resource from a json description passed on stdin.\n\nUsage:
     exo [options] update <cik> [<rid>] -
@@ -126,18 +136,21 @@ Dispatch Description
     exo [options] lookup <cik> --cik=<cik-to-find>''',
     'drop':
         '''Drop (permanently delete) a resource.\n\nUsage:
-    exo [options] drop <cik> [<rid> ...]''',
+    exo [options] drop <cik> [<rid> ...]
+
+Options:
+    --all-children  drop all children of the resource.''',
     'flush':
         '''Remove all time series data from a resource.\n\nUsage:
     exo [options] flush <cik> [<rid>]''',
     'tree':
         '''Display a resource's descendants.\n\nUsage:
     exo tree [--verbose] [--hide-keys] <cik>''',
-    'drop-all-children':
-        '''Drop (delete permanently) all children of a resource.\n\nUsage:
-    exo [options] drop-all-children <cik>''',
     'script': '''Upload a Lua script\n\nUsage:
-    exo [options] script <script-file> <cik> [--name=<name>]'''
+    exo [options] script <script-file> <cik> ...
+
+Options:
+    --name=<name>  script name, if different from script filename.'''
 }
 
 for k in cmd_doc:
@@ -199,7 +212,9 @@ class ExoRPC():
         isok, response = self.exo.record(cik, rid, entries, {})
         self._raise_for_response_record(isok, response)
 
-    def create(self, cik, type, desc):
+    def create(self, cik, type, desc, name=None):
+        if name is not None:
+            desc['name'] = name
         isok, response = self.exo.create(cik, type, desc)
         self._raise_for_response(isok, response)
         return response
@@ -223,10 +238,10 @@ class ExoRPC():
         if desc is None:
             # default description
             desc = {'limits': {'dataport': 'inherit',
-                                'datarule': 'inherit',
-                                'dispatch': 'inherit',
-                                'disk': 'inherit',
-                                'io': 'inherit'},
+                               'datarule': 'inherit',
+                               'dispatch': 'inherit',
+                               'disk': 'inherit',
+                               'io': 'inherit'},
                     'writeinterval': 'inherit'}
         if name is not None:
             desc['name'] = name
@@ -344,14 +359,22 @@ class ExoRPC():
         def add_opt(o, label, value):
             if o is True or (o in cli_args and cli_args[o] is True):
                 opt[label] = value
-        add_opt(True, 'type', typ)
+        add_opt(True, typ + ' name', name)
         if 'format' in info['description']:
             add_opt(True, 'format', info['description']['format'])
-        add_opt(True, 'name', name)
-        add_opt(True, 'aliases', str(aliases))
-        add_opt('--verbose', 'unit', units)
+
+        has_alias = aliases is not None and len(aliases) > 0
+        if has_alias:
+            add_opt(True, 'aliases', str(aliases))
+        # show RID for clients with no alias, or if --verbose was passed
+        ridopt = False
         if typ == 'client':
-            add_opt('--verbose', 'rid', self._disp_key(cli_args, rid))
+            if has_alias:
+                ridopt = '--verbose'
+            else:
+                ridopt = True
+        add_opt(ridopt, 'rid', self._disp_key(cli_args, rid))
+        add_opt('--verbose', 'unit', units)
         if 'storage' in info and 'count' in info['storage']:
             add_opt(True, 'count', info['storage']['count'])
 
@@ -372,6 +395,7 @@ class ExoRPC():
             info = self.info(cik, {"alias": ""})
             # info doesn't contain key
             info['key'] = cik
+            aliases = info['aliases']
             root_aliases = '(see parent)'
             # todo: can I get aliases for cik? For now, pass []
             self._print_node(rid,
@@ -469,19 +493,20 @@ class ExoRPC():
             else:
                 raise ExoException("Error updating datarule.")
 
-    def upload(self, cik, filename, name=None):
+    def upload(self, ciks, filename, name=None):
         try:
             f = open(filename)
         except IOError:
-            raise ExoException('Error opening file.')
+            raise ExoException('Error opening file {}.'.format(filename))
         else:
             with f:
                 text = f.read().strip()
                 if name is None:
                     # if no name is specified, use the file name as a name
                     name = os.path.basename(filename)
-                rid = self._lookup_rid_by_name(cik, name)
-                self._upload_script(cik, name, text, rid)
+                for cik in ciks:
+                    rid = self._lookup_rid_by_name(cik, name)
+                    self._upload_script(cik, name, text, rid)
 
     def lookup_rid(self, cik, cik_to_find):
         isok, listing = self.exo.listing(cik, types=['client'])
@@ -500,10 +525,14 @@ class ExoRPC():
                     return listing[0][idx]
 
     def record_backdate(self, cik, rid, interval_seconds, values):
-        '''Record a list of values and record them as if they happened in the past
-        interval_seconds apart. For example, if values ['a', 'b', 'c'] are passed in
-        with interval 10, they're recorded as [[0, 'c'], [-10, 'b'], [-20, 'a']]'''
+        '''Record a list of values and record them as if they happened in
+        the past interval_seconds apart. For example, if values
+            ['a', 'b', 'c']
+        are passed in with interval 10, they're recorded as
+            [[0, 'c'], [-10, 'b'], [-20, 'a']].
+        interval_seconds must be positive.'''
         timestamp = -interval_seconds
+
         tvalues = []
         values.reverse()
         for v in values:
@@ -512,6 +541,165 @@ class ExoRPC():
         return self.record(cik, rid, tvalues)
 
 
+def format_time(sec):
+    '''Formats a time interval for human consumption'''
+    intervals = [[60 * 60 * 24, 'd'],
+                    [60 * 60, 'h'],
+                    [60, 'm']]
+    text = ""
+    for s, label in intervals:
+        if sec > s:
+            text = "{} {}{}".format(text, sec / s, label)
+            sec -= s * (sec / s)
+    return "{} {}s".format(text, sec).strip()
+
+
+def spark(numbers, empty_val=None):
+    """Generate a text based sparkline graph from a list of numbers (ints or
+    floats).
+
+    When value is empty_val, show no bar.
+
+    https://github.com/1stvamp/py-sparkblocks
+
+    Based on:
+      https://github.com/holman/spark
+    and:
+      http://www.datadrivenconsulting.com/2010/06/twitter-sparkline-generator/
+    """
+
+    out = []
+
+    min_value = min(numbers)
+    max_value = max(numbers)
+    value_scale = max_value - min_value
+
+    for number in numbers:
+        if number == empty_val:
+            out.append(u" ")
+        else:
+            if (number - min_value) != 0 and value_scale != 0:
+                scaled_value = (number - min_value) / value_scale
+            else:
+                scaled_value = 0
+            num = math.floor(min([6, (scaled_value * 7)]))
+
+            # Hack because 9604 and 9608 aren't vertically aligned the same as
+            # other block elements
+            if num == 3:
+                if (scaled_value * 7) < 3.5:
+                    num = 2
+                else:
+                    num = 4
+            elif num == 7:
+                num = 6
+
+            out.append(unichr(int(9601 + num)))
+
+    return ''.join(out)
+
+def show_intervals(er, cik, rid, limit):
+    # show a distribution of intervals between data
+    data = er.read(cik,
+                   rid,
+                   sort='desc',
+                   limit=limit)
+
+    intervals = [data[i - 1][0] - data[i][0] for i in xrange(1, len(data))]
+    intervals = sorted(intervals)
+    num_bins = 60
+    min_t, max_t = min(intervals), max(intervals)
+    bin_size = float(max_t - min_t) / num_bins * 1.0
+
+    bins = []
+    for i in range(num_bins):
+        bin_min = min_t + i * bin_size
+        bin_max = min_t + (i + 1) * bin_size
+        if i != 0:
+            critfn = lambda x: bin_min < x and x <= bin_max
+        else:
+            critfn = lambda x: bin_min <= x and x <= bin_max
+        #bins.append((bin_min, bin_max, float(
+        #    sum(itertools.imap(critfn, intervals)))))
+        bins.append(float(
+            sum(itertools.imap(critfn, intervals))))
+
+    if False:
+        # debug
+        print(bins)
+        print(num_bins * bin_size)
+        print(bin_size)
+        print(min_t)
+        print(max_t)
+        print(set(intervals))
+
+    print(spark(bins, empty_val=0))
+
+    min_label = format_time(min_t)
+    max_label = format_time(max_t)
+    sys.stdout.write(min_label)
+    sys.stdout.write(' ' * (num_bins - len(min_label) - len(max_label)))
+    sys.stdout.write(max_label + '\n')
+
+def read_cmd(er, cik, rids, args):
+    rid = rids[0]
+    limit = args['--limit']
+    limit = 1 if limit is None else int(limit)
+
+    # --intervals is practically another command. Only
+    # shares time specifications (see above).
+    if args['--intervals']:
+        show_intervals(er, cik, rid, limit)
+        return
+
+    timeformat = args['--timeformat']
+    dw = csv.DictWriter(sys.stdout, ['timestamp', 'value'])
+    fmt = args['--format']
+
+    def printline(timestamp, val):
+        if fmt == 'raw':
+            print(val)
+        else:
+            if timeformat == 'unix':
+                dt = timestamp
+            else:
+                dt = datetime.fromtimestamp(timestamp)
+            dw.writerow({'timestamp': str(dt), 'value': val})
+
+    sleep_seconds = 2
+    if args['--follow']:
+        results = []
+        while len(results) == 0:
+            results = er.read(cik,
+                                rid,
+                                limit=1,
+                                sort='desc')
+            if len(results) > 0:
+                last_t, last_v = results[0]
+                printline(last_t, last_v)
+            else:
+                time.sleep(sleep_seconds)
+
+        while True:
+            results = er.read(cik,
+                                rid,
+                                limit=10000,
+                                starttime=last_t + 1)
+
+            for t, v in results:
+                printline(t, v)
+
+            if len(results) > 0:
+                last_t, last_v = results[-1]
+
+            time.sleep(sleep_seconds)
+    else:
+        for t, v in er.read(cik,
+                            rid,
+                            sort='desc',
+                            limit=limit):
+            printline(t, v)
+
 def plain_print(arg):
     print(arg)
 
@@ -544,57 +732,12 @@ def handle_args(cmd, args):
         pr = plain_print
 
     if cmd == 'read':
-        rid = rids[0]
-        limit = args['--limit']
-        limit = 1 if limit is None else int(limit)
-        dr = csv.DictWriter(sys.stdout, ['timestamp', 'value'])
-        fmt = args['--format']
-
-        def printline(timestamp, val):
-            if fmt == 'raw':
-                print(val)
-            else:
-                dt = datetime.fromtimestamp(timestamp)
-                dr.writerow({'timestamp': str(dt), 'value': val})
-
-        sleep_seconds = 2
-        if args['--follow']:
-            results = []
-            while len(results) == 0:
-                results = er.read(cik,
-                                  rid,
-                                  limit=1,
-                                  sort='desc')
-                if len(results) > 0:
-                    last_t, last_v = results[0]
-                    printline(last_t, last_v)
-                else:
-                    time.sleep(sleep_seconds)
-
-            while True:
-                results = er.read(cik,
-                                  rid,
-                                  limit=10000,
-                                  starttime=last_t + 1)
-
-                for t, v in results:
-                    printline(t, v)
-
-                if len(results) > 0:
-                    last_t, last_v = results[-1]
-
-                time.sleep(sleep_seconds)
-        else:
-            for t, v in er.read(cik,
-                                rid,
-                                sort='desc',
-                                limit=limit):
-                printline(t, v)
+        read_cmd(er, cik, rids, args)
     elif cmd == 'write':
         er.write(cik, rids[0], args['--value'])
     elif cmd == 'record':
         interval = args['--interval']
-        if interval is not None:
+        if interval is None:
             entries = []
             # split timestamp, value
             if args['-']:
@@ -623,16 +766,44 @@ def handle_args(cmd, args):
                 values = [v.strip() for v in sys.stdin.readlines()]
             else:
                 values = args['--value']
-            er.record_backdate(cik, rids[0], int(interval), values)
-
-
+            interval = int(interval)
+            if interval <= 0:
+                raise ExoException("--interval must be positive")
+            er.record_backdate(cik, rids[0], interval, values)
     elif cmd == 'create':
-        s = sys.stdin.read()
-        try:
-            desc = json.loads(s)
-        except Exception as ex:
-            raise ExoException(ex)
-        pr(er.create(cik, type=args['--type'], desc=desc))
+        typ = args['--type']
+        if args['-']:
+            s = sys.stdin.read()
+            try:
+                desc = json.loads(s)
+            except Exception as ex:
+                raise ExoException(ex)
+            rid = er.create(cik,
+                            type=typ,
+                            desc=desc,
+                            name=args['--name'])
+        elif typ == 'client':
+            rid = er.create_client(cik,
+                                   name=args['--name'])
+        elif typ == 'dataport':
+            rid = er.create_dataport(cik,
+                                     args['--format'],
+                                     name=args['--name'])
+        else:
+            raise ExoException('No defaults for {}.'.format(args['--type']))
+        ridonly = args['--ridonly']
+        if ridonly:
+            pr(rid)
+        else:
+            pr('rid: {}'.format(rid))
+        if not ridonly and  typ == 'client':
+            # for convenience, look up the cik
+            print('cik: {}'.format(er.info(cik, rid, cikonly=True)))
+        if args['--alias'] is not None:
+            er.map(cik, rid, args['--alias'])
+            if not ridonly:
+                print("alias: {}".format(args['--alias']))
+
     elif cmd == 'update':
         s = sys.stdin.read()
         try:
@@ -640,10 +811,6 @@ def handle_args(cmd, args):
         except Exception as ex:
             raise ExoException(ex)
         pr(er.update(cik, rids[0], desc=desc))
-    elif cmd == 'create-dataport':
-        pr(er.create_dataport(cik, args['--format'], name=args['--name']))
-    elif cmd == 'create-client':
-        pr(er.create_client(cik, name=args['--name']))
     elif cmd == 'map':
         er.map(cik, rids[0], args['<alias>'])
     elif cmd == 'unmap':
@@ -657,11 +824,14 @@ def handle_args(cmd, args):
         else:
             pr(er.lookup(cik, args['<alias>']))
     elif cmd == 'drop':
-        er.drop(cik, rids)
+        if args['--all-children']:
+            er.drop_all_children(cik)
+        else:
+            er.drop(cik, rids)
     elif cmd == 'listing':
         types = args['--type']
         listing = er.listing(cik, types)
-        if args['--plain'] == True:
+        if args['--plain']:
             for l in listing:
                 for cik in listing[0]:
                     print(cik)
@@ -679,17 +849,32 @@ def handle_args(cmd, args):
     # special commands
     elif cmd == 'tree':
         er.tree(cik, cli_args=args)
-    elif cmd == 'drop-all-children':
-        er.drop_all_children(cik)
     elif cmd == 'script':
+        # cik is a list of ciks
         er.upload(cik, args['<script-file>'], args['--name'])
     else:
         raise ExoException("Command not handled")
 
-if __name__ == '__main__':
 
-    args = docopt(__doc__,
-                  version="Exosite RPC API Command Line {}".format(__version__), options_first=True)
+
+def cmd(argv=None, stdin=None, stdout=None, stderr=None):
+    '''Wrap the command line interface. Globally redirects args
+    and io so that the application can be tested externally.'''
+
+    # globally redirect args and io
+    if argv is not None:
+        sys.argv = argv
+    if stdin is not None:
+        sys.stdin = stdin
+    if stderr is not None:
+        sys.stderr = stderr
+    if stdout is not None:
+        sys.stdout = stdout
+
+    args = docopt(
+        __doc__,
+        version="Exosite RPC API Command Line {}".format(__version__),
+        options_first=True)
 
     # get command args
     cmd = args['<command>']
@@ -698,7 +883,7 @@ if __name__ == '__main__':
         args_cmd = docopt(cmd_doc[cmd], argv=argv)
     else:
         print('Unknown command {}. Try "exo --help"'.format(cmd))
-        sys.exit(1)
+        return 1
 
     # merge command-specific arguments into general arguments
     args.update(args_cmd)
@@ -712,15 +897,20 @@ if __name__ == '__main__':
     except ExoException as ex:
         # command line tool threw an exception on purpose
         sys.stderr.write("Command line error: {}\r\n".format(ex))
-        sys.exit(1)
+        return 1
     except RPCException as ex:
         # pyonep library call signaled an error in return values
         sys.stderr.write("One Platform error: {}\r\n".format(ex))
-        sys.exit(1)
+        return 1
     except onep_exceptions.OnePlatformException as ex:
         # pyonep library call threw an exception on purpose
         sys.stderr.write("One Platform exception: {}\r\n".format(ex))
-        sys.exit(1)
+        return 1
     except KeyboardInterrupt:
         if args['--debug']:
             raise
+    return 0
+
+
+if __name__ == '__main__':
+    sys.exit(cmd(sys.argv))
